----------------------------------------
-- _Clock
----------------------------------------

GroupCalendar._Clock = {}

function GroupCalendar._Clock:_New(pSize, pFrame)
	if pFrame then
		return pFrame
	end
	
	return CreateFrame("Frame", nil, UIParent)
end

function GroupCalendar._Clock:_Construct(pSize, pFrame, pOffsetX, pOffsetY, pShowDate)
	if not pSize then
		pSize = 22
	end
	
	if not pFrame then
		self:SetWidth(pSize)
		self:SetHeight(pSize)
	end
	
	self.ClockBackground = self:CreateTexture(nil, "BORDER")
	self.ClockBackground:SetTexture(GroupCalendar.UI.AddonPath.."Textures\\ClockBackground")
	self.ClockBackground:SetPoint("CENTER", self, "CENTER", pOffsetX, pOffsetY)
	self.ClockBackground:SetWidth(pSize)
	self.ClockBackground:SetHeight(pSize)
	
	self.MinuteHand = self:CreateTexture(nil, "ARTWORK")
	self.MinuteHand:SetTexture(GroupCalendar.UI.AddonPath.."Textures\\ClockHand")
	self.MinuteHand:SetPoint("CENTER", self, "CENTER", pOffsetX, pOffsetY)
	self.MinuteHand:SetWidth(pSize - 10)
	self.MinuteHand:SetHeight(pSize - 10)
	
	self.HourHand = self:CreateTexture(nil, "ARTWORK")
	self.HourHand:SetTexture(GroupCalendar.UI.AddonPath.."Textures\\ClockHourHand")
	self.HourHand:SetPoint("CENTER", self, "CENTER", pOffsetX, pOffsetY)
	self.HourHand:SetWidth(pSize - 10)
	self.HourHand:SetHeight(pSize - 10)
	
	self.ClockGloss = self:CreateTexture(nil, "OVERLAY")
	self.ClockGloss:SetTexture(GroupCalendar.UI.AddonPath.."Textures\\ClockGloss")
	self.ClockGloss:SetPoint("CENTER", self, "CENTER", pOffsetX, pOffsetY)
	self.ClockGloss:SetWidth(pSize)
	self.ClockGloss:SetHeight(pSize)
	
	if pShowDate then
		self:SetText("") -- Clear any existing text
		self.SetText = function (self, pText) self.DateFontString:SetText(pText) end
		
		self.DateFontString = self:CreateFontString(nil, "ARTWORK", "SystemFont_Small")
		self.DateFontString:SetTextColor(77 / 255, 56 / 255, 0 / 255, 0.75)
		self.DateFontString:SetPoint("TOP", self, "CENTER", -2, 0)
	end
	
	self.NextUpdateDelay = 0
	
	self:HookScript("OnShow", function () self:OnShow() end)
	self:HookScript("OnHide", function () self:OnHide() end)
	self:HookScript("OnUpdate", function (pFrame, pElapsed) self:OnUpdate(pElapsed) end)
	
	if self:IsVisible() then
		self:OnShow()
	end
end

function GroupCalendar._Clock:OnShow()
	self.NextUpdateDelay = 0 -- Force an update
	self:OnUpdate(0)
	
	if GroupCalendar.EventLib then
		GroupCalendar.EventLib:RegisterEvent("GC_CLOCKS_CHANGED", self.Update, self)
		GroupCalendar.EventLib:RegisterEvent("GC5_PREFS_CHANGED", function () self.NextUpdateDelay = 0 end)
	end
end

function GroupCalendar._Clock:OnHide()
	if GroupCalendar.EventLib then
		GroupCalendar.EventLib:UnregisterEvent("GC_CLOCKS_CHANGED", self.Update, self)
		GroupCalendar.EventLib:UnregisterEvent("GC5_PREFS_CHANGED", nil, self)
	end
end

function GroupCalendar._Clock:SetShowLocalTime(pShowLocalTime)
	GroupCalendar.Clock.Data.ShowLocalTime = pShowLocalTime
	
	if GroupCalendar.EventLib then
		GroupCalendar.EventLib:DispatchEvent("GC5_PREFS_CHANGED")
	end
end

function GroupCalendar._Clock:OnUpdate(pElapsed)
	self.NextUpdateDelay = self.NextUpdateDelay - pElapsed
	
	if self.NextUpdateDelay > 0 then
		return
	end
	
	self.NextUpdateDelay = 30
	
	self:Update()
end

function GroupCalendar._Clock:Update()
	local vHour, vMinute = GetGameTime()
	local vHourAngle
	
	if GroupCalendar.Clock.Data.ShowLocalTime then
		vHour, vMinute = GroupCalendar.DateLib:ConvertTimeToHM(GroupCalendar.DateLib:GetLocalTimeFromServerTime(GroupCalendar.DateLib:ConvertHMToTime(vHour, vMinute)))
	end
	
	vHourAngle = (vHour + vMinute / 60) * 3.1415926535 / 6
	
	self:SetTextureAngle(self.HourHand, vHourAngle)
	self:SetTextureAngle(self.MinuteHand, vMinute * 3.1415926535 / 30)
	
	if vHour ~= self.PreviousHour then
		self.PreviousHour = vHour
		GameTimeFrame_SetDate() -- Refresh the date each hour
	end
end

function GroupCalendar._Clock:MatrixDotVector(pMatrix, pVector)
	local vResult = {}
	
	for vRow, vRowValues in ipairs(pMatrix) do
		local vTotal = 0
		
		for vColumn, vMatrixValue in ipairs(vRowValues) do
			vTotal = vTotal + vMatrixValue * (pVector[vColumn] or 1)
		end
		
		vResult[vRow] = vTotal
	end
	
	return vResult
end

function GroupCalendar._Clock:SetTextureAngle(pTexture, pAngle, pScaleX, pScaleY)
	-- Calculate the rotation transform
	
	local vCosAngle = math.cos(-pAngle)
	local vSinAngle = math.sin(-pAngle)
	
	local vTransform =
	{
		{vCosAngle, vSinAngle, 0.5}, -- Offset by 0.5 to make the coordinates 0.0 to 1.0 instead of -0.5 to 0.5
		{vSinAngle, -vCosAngle, 0.5},-- Same for the Y axis
	}
	
	-- Rotate the texture
	
	local vTopLeft = self:MatrixDotVector(vTransform, {-0.5, 0.5})
	local vTopRight = self:MatrixDotVector(vTransform, {0.5, 0.5})
	local vBottomLeft = self:MatrixDotVector(vTransform, {-0.5, -0.5})
	local vBottomRight = self:MatrixDotVector(vTransform, {0.5, -0.5})
	
	-- Set the texture
	
	pTexture:SetTexCoord(
			vTopLeft[1], vTopLeft[2],
			vBottomLeft[1], vBottomLeft[2],
			vTopRight[1], vTopRight[2],
			vBottomRight[1], vBottomRight[2])
end
